#!/usr/bin/env python3
# @nolint
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2024 by Rivos Inc.
# Licensed under the Apache License, Version 2.0, see LICENSE for details.
# SPDX-License-Identifier: Apache-2.0

from generator_common import (
    args,
    clang_format,
    extract_node,
    parse,
    read_extent,
    to_snake_case,
    CursorKind,
    COPYRIGHT,
    FUNCTION_CURSOR_KINDS,
)

import io
import subprocess


AUTOGEN_HEADER = f"""/*
 * This file is auto-generated from test_fixture_generator.py
 * DO NOT EDIT!
 */
"""


def get_children_with_type(cursor, ty):
    return list(filter(lambda c: c.kind == ty, cursor.get_children()))


def get_child_with_type(cursor, ty):
    children = get_children_with_type(cursor, ty)
    assert len(children) == 1
    return children[0]


class TestFixtureGen:
    def includes(self, fixture_name):
        raise NotImplementedError

    def to_kernel_args(self, name, ty, kernel_args, preamble, postamble):
        if "std::vector" in ty:
            kernel_args.append(name)
        elif "*" in ty:
            # remove ptr from ty
            dty = ty[:-2]
            preamble.append(f"std::vector<{dty}> vec_{name}(1, *{name});")
            postamble.append(f"*{name} = vec_{name}[0];")
            kernel_args.append(f"vec_{name}")
        else:
            preamble.append(f"const std::vector<{ty}> vec_{name}(1, {name});")
            kernel_args.append(f"vec_{name}")

    def body(self, method, fixture_type_param, template_params, function_params):
        raise NotImplementedError

    def methods(self, test_fixture, fixture_type_param):
        methods = []
        for method in filter(
            lambda c: c.kind in FUNCTION_CURSOR_KINDS, test_fixture["children"]
        ):
            method = extract_node(method)

            template_params = list(
                map(extract_node, method["fn_props"]["template_params"])
            )
            function_params = list(
                map(extract_node, method["fn_props"]["function_params"])
            )

            templates = ", ".join(
                map(lambda node: read_extent(node["cursor"].extent), template_params)
            )
            templates = f"template <{templates}>" if len(templates) != 0 else ""
            name = method["spelling"]
            params = ", ".join(
                map(lambda node: read_extent(node["cursor"].extent), function_params)
            )
            body = self.body(
                method, fixture_type_param, template_params, function_params
            )

            methods.append(
                f"""
            {templates}
            void {name}({params}) {{
                {body}
            }}"""
            )

        return methods

    def class_preamble(self, fixture_name):
        return ""

    def generate(self, tu, filename):
        out = io.StringIO()

        test_fixture = extract_node(
            get_child_with_type(tu.cursor, CursorKind.CLASS_TEMPLATE)
        )
        assert (
            get_child_with_type(
                test_fixture["cursor"], CursorKind.CXX_BASE_SPECIFIER
            ).spelling
            == "::testing::Test"
        )

        fixture_name = test_fixture["spelling"]
        fixture_type_param = get_child_with_type(
            test_fixture["cursor"], CursorKind.TEMPLATE_TYPE_PARAMETER
        ).spelling
        methods = "\n".join(self.methods(test_fixture, fixture_type_param))

        out.write(
            f"""\
        {COPYRIGHT}

        {AUTOGEN_HEADER}

        #include <gtest/gtest.h>
        #include <vector>

        {self.includes(fixture_name)}

        template <typename {fixture_type_param}>
        class {fixture_name} : public ::testing::Test {{
         {self.class_preamble(fixture_name)}

         protected:
            {methods}
        }};
        """
        )

        with open(filename, "w") as outf:
            outf.write(out.getvalue())
        subprocess.run([clang_format, "-i", filename])


class HipTestFixture(TestFixtureGen):
    def __init__(self):
        self.launcher_fn = "HipTestLaunch"

    def includes(self, fixture_name):
        test_type = fixture_name.replace("Test", "").lower()
        return f"""
            #include <hip/hip_runtime.h>
            #include "test/generated/{test_type}s/kernels-hip.hpp"
            #include "test/platforms/hip_test.hpp"
        """

    def needs_shared_mem(self):
        return False

    def get_kernel_template_args(self, method, fixture_type_param, template_params):
        # Whether the kernel actually depends on the fixture type parameter
        untyped = method["fn_props"]["attrs"].get("untyped", False)
        if untyped:
            return list(map(lambda node: node["spelling"], template_params))

        templates = []
        inserted_fixture_type = False
        for t in template_params:
            # FIXME: The fixture template parameter (usually T) should always
            # be inserted before `U` or `H` (if its being used)
            if t["spelling"] in ["U", "H"]:
                templates.append(fixture_type_param)
                templates.append(t["spelling"])
                inserted_fixture_type = True
            else:
                # FIXME: No need to pass the template template parameter instead of just the
                # fully specified type in the tests.
                templates.append(
                    t["spelling"]
                    if not t["kind"] == CursorKind.TEMPLATE_TEMPLATE_PARAMETER
                    else t["spelling"] + f"<{fixture_type_param}>"
                )
        if not inserted_fixture_type:
            templates.append(fixture_type_param)

        return templates

    def body(self, method, fixture_type_param, template_params, function_params):
        preamble = []
        postamble = []
        maybe_add_shared_mem = ""

        if self.needs_shared_mem():
            shared_mem_type = method["fn_props"]["attrs"].get("shared_mem_type")
            if shared_mem_type:
                assert len(shared_mem_type) <= 1
                # Only OpenMP uses this code path and it requires that WARP_THREADS == BLOCK_THREADS
                shared_mem_type = shared_mem_type[0].replace(
                    "WARP_THREADS", "/*WARP_THREADS=*/BLOCK_THREADS"
                )
                platformt = ""
                if "PlatformT" in shared_mem_type:
                    platformt = f"using PlatformT = {self.platform_type}<BLOCK_THREADS, /*WARP_THREADS=*/BLOCK_THREADS>;"
                preamble += [
                    f"""
                    {platformt}
                    using SharedMemType = {shared_mem_type};"""
                ]
                maybe_add_shared_mem = ", SharedMemType"

        templates = ", ".join(
            self.get_kernel_template_args(method, fixture_type_param, template_params)
        )
        templates += maybe_add_shared_mem
        kernel_template_args = f"<{templates}>" if len(templates) != 0 else ""

        thread_count = (
            "BLOCK_THREADS"
            if "BLOCK_THREADS" in kernel_template_args
            else "/*BLOCK_THREADS=*/1"
        )
        block_count = "/*num_blocks=*/1"
        kernel_args = []
        last_arg = ""
        for p in function_params:
            name = p["spelling"]
            ty = p["type_spelling"]
            attrs = list(
                map(
                    lambda c: c.spelling,
                    get_children_with_type(p["cursor"], CursorKind.ANNOTATE_ATTR),
                )
            )
            assert len(attrs) <= 1
            if "use_as_size" in attrs:
                last_arg = name + ".size()"
            if "block_count" in attrs:
                block_count = name
                continue

            self.to_kernel_args(name, ty, kernel_args, preamble, postamble)

        if len(last_arg) != 0:
            kernel_args.append(last_arg)

        preamble = "".join(preamble)
        postamble = "".join(postamble)
        kernel_args = ", ".join([""] + kernel_args) if kernel_args else ""
        return f"""
            {preamble}
            {self.launcher_fn}<{thread_count}{maybe_add_shared_mem}>(
                {block_count},
                &kernels::{method['spelling']}{kernel_template_args}{kernel_args}
            );
            {postamble}"""


class CudaTestFixture(HipTestFixture):
    def __init__(self):
        self.launcher_fn = "CudaTestLaunch"

    def includes(self, fixture_name):
        test_type = fixture_name.replace("Test", "").lower()
        return f"""
            #include "test/generated/{test_type}s/kernels-cuda.cuh"
            #include "breeze/platforms/cuda.cuh"
            #include "test/platforms/cuda_test.cuh"
        """


class OpenmpTestFixture(HipTestFixture):
    def __init__(self):
        self.launcher_fn = "OpenMPTestLaunch"
        self.platform_type = "OpenMPPlatform"

    def includes(self, fixture_name):
        test_type = fixture_name.replace("Test", "").lower()
        return f"""
            #include <omp.h>
            #include <cassert>

            #include "test/generated/{test_type}s/kernels-openmp.h"
            #include "breeze/platforms/openmp.h"
            #include "test/platforms/openmp_test.h"
        """

    def needs_shared_mem(self):
        return True

    def to_kernel_args(self, name, ty, kernel_args, preamble, postamble):
        if "std::vector" in ty:
            kernel_args.append(f"{name}.data()")
        elif "*" in ty:
            kernel_args.append(name)
        else:
            kernel_args.append(f"&{name}")


class SyclTestFixture(HipTestFixture):
    def __init__(self):
        self.launcher_fn = "SyclTestSubmit"
        self.platform_type = "SyCLPlatform"

    def includes(self, fixture_name):
        test_type = fixture_name.replace("Test", "").lower()
        return f"""
            #pragma GCC diagnostic push
            #pragma GCC diagnostic ignored "-Wunused-parameter"
            #pragma GCC diagnostic ignored "-Wunused-local-typedef"
            #pragma GCC diagnostic ignored "-Wunused-variable"
            #pragma GCC diagnostic ignored "-Wsign-compare"
            #pragma GCC diagnostic ignored "-Wmismatched-tags"
            #pragma GCC diagnostic ignored "-Wdeprecated-copy"
            #include <CL/sycl.hpp>
            #pragma GCC diagnostic pop

            #include "test/generated/{test_type}s/kernels-sycl.hpp"
            #include "test/platforms/sycl_test.hpp"

            using kernels::WARP_THREADS;
        """

    def needs_shared_mem(self):
        return True

    # Shares the same overall structure as HIP and OpenMP, but more work needs
    # to be done to first create a wrapping lambda to pass into SyclTestSubmit
    def body(self, method, fixture_type_param, template_params, function_params):
        preamble = []
        postamble = []
        maybe_add_shared_mem = ""
        last_inner_args = []
        last_lambda_params = []

        if self.needs_shared_mem():
            shared_mem_type = method["fn_props"]["attrs"].get("shared_mem_type")
            if shared_mem_type:
                assert len(shared_mem_type) <= 1
                shared_mem_type = shared_mem_type[0]
                platformt = ""
                if "PlatformT" in shared_mem_type:
                    platformt = f"using PlatformT = {self.platform_type}<BLOCK_THREADS, WARP_THREADS>;"
                preamble += [
                    f"""
                    {platformt}
                    using SharedMemType = {shared_mem_type};"""
                ]
                maybe_add_shared_mem = ", SharedMemType"
                last_inner_args += ["shared_mem"]
                last_lambda_params += ["SharedMemType* shared_mem"]

        templates = ", ".join(
            self.get_kernel_template_args(method, fixture_type_param, template_params)
        )
        templates += maybe_add_shared_mem
        kernel_template_args = f"<{templates}>" if len(templates) != 0 else ""

        thread_count = (
            "BLOCK_THREADS"
            if "BLOCK_THREADS" in kernel_template_args
            else "/*BLOCK_THREADS=*/1"
        )
        block_count = "/*num_blocks=*/1"
        inner_kernel_args = ["item"]
        kernel_args = []
        lambda_params = ["cl::sycl::nd_item<1> item"]
        size_arg = ""
        for p in function_params:
            name = p["spelling"]
            ty = p["type_spelling"]
            attrs = list(
                map(
                    lambda c: c.spelling,
                    get_children_with_type(p["cursor"], CursorKind.ANNOTATE_ATTR),
                )
            )
            assert len(attrs) <= 1
            if "use_as_size" in attrs:
                size_arg = "num_items = " + name + ".size()"
                last_inner_args += ["num_items"]
            if "block_count" in attrs:
                block_count = name
                continue

            self.to_kernel_args(name, ty, kernel_args, preamble, postamble)

            # All arguments to SyclTestSubmit are either already vectors or have
            # now been copied to a vector, and so for the respective lambda
            # params we can decay to the underlying pointer type which should
            # handle any const-ness as appropriate.
            lambda_params.append(f"decltype({kernel_args[-1]}.data()) {name}")
            inner_kernel_args += [name]

        inner_kernel_args += last_inner_args
        lambda_params += last_lambda_params

        preamble = "".join(preamble)
        postamble = "".join(postamble)
        kernel_args = ", ".join([""] + kernel_args) if kernel_args else ""
        inner_kernel_args = ", ".join(inner_kernel_args)
        lambda_params = ", ".join(lambda_params)
        lambda_fn = f"""\
            [{size_arg}]({lambda_params}){{
                kernels::{method['spelling']}{kernel_template_args}({inner_kernel_args});
            }}\
        """
        return f"""
            {preamble}
            {self.launcher_fn}<{thread_count}{maybe_add_shared_mem}>(
                {block_count}, {lambda_fn}{kernel_args}
            );
            {postamble}"""


class OpenclTestFixture(TestFixtureGen):
    def __init__(self):
        self.launcher_fn = "OpenCLTestDispatch"

    def includes(self, fixture_name):
        return f"""
            #include <memory>

            #include "test/platforms/opencl_test.h"
            #include "test/type_helpers.h"
        """

    def class_preamble(self, fixture_name):
        return f"""
         private:
          struct Impl {{
            OpenCLTest test;
          }};
          std::unique_ptr<Impl> impl_;

         protected:
          {fixture_name}() : impl_(std::make_unique<Impl>()) {{}}
          ~{fixture_name}() override {{}}

          void SetUp() override {{}}
          void TearDown() override {{}}
        """

    def get_kernel_template_args(self, method, fixture_type_param, template_params):
        template_variants = []
        template_sizes = []

        for t in template_params:
            if t["kind"] == CursorKind.TEMPLATE_TYPE_PARAMETER:
                template_variants += [t["spelling"]]
            elif t["kind"] == CursorKind.TEMPLATE_TEMPLATE_PARAMETER:
                template_variants += [t["spelling"] + f"<{fixture_type_param}>"]
            elif t["kind"] == CursorKind.TEMPLATE_NON_TYPE_PARAMETER:
                template_sizes += [t["spelling"]]

            # FIXME: No need to pass the template template parameter instead of just the
            # fully specified type in the tests.

        # Whether the kernel actually depends on the fixture type parameter
        untyped = method["fn_props"]["attrs"].get("untyped", False)
        if not untyped:
            template_variants += [fixture_type_param]

        return (template_variants, template_sizes)

    def framework_specific_args(self):
        return "&impl_->test, kernel_name"

    def body(self, method, fixture_type_param, template_params, function_params):
        preamble = []
        postamble = []

        (template_variants, template_sizes) = self.get_kernel_template_args(
            method, fixture_type_param, template_params
        )
        op = "op"
        if template_variants:
            op = f"add_op_variant<{op}, {', '.join(template_variants)}>"
        if template_sizes:
            op = f"add_instance_shape<{op}, {', '.join(template_sizes)}>"

        thread_count = (
            "BLOCK_THREADS"
            if "BLOCK_THREADS" in template_sizes
            else "/*BLOCK_THREADS=*/1"
        )
        block_count = "/*numBlocks=*/1"
        kernel_args = []
        last_arg = ""
        for p in function_params:
            name = p["spelling"]
            ty = p["type_spelling"]
            attrs = list(
                map(
                    lambda c: c.spelling,
                    get_children_with_type(p["cursor"], CursorKind.ANNOTATE_ATTR),
                )
            )
            assert len(attrs) <= 1
            if "use_as_size" in attrs:
                last_arg = f"static_cast<int>({name}.size())"
            if "block_count" in attrs:
                block_count = name
                continue

            self.to_kernel_args(name, ty, kernel_args, preamble, postamble)

        if len(last_arg) != 0:
            kernel_args.append(last_arg)

        preamble += [
            f'constexpr static const char op[] = "{to_snake_case(method["spelling"])}";'
        ]
        preamble += [f"constexpr const char* kernel_name = {op};"]

        preamble = "".join(preamble)
        postamble = "".join(postamble)
        kernel_args = ", ".join([""] + kernel_args) if kernel_args else ""
        return f"""
            {preamble}
            {self.launcher_fn}<{thread_count}>(
                {block_count}, {self.framework_specific_args()}{kernel_args}
            );
            {postamble}"""


class MetalTestFixture(OpenclTestFixture):
    def __init__(self):
        self.launcher_fn = "MetalTestDispatch"

    def includes(self, fixture_name):
        return f"""
            #include <memory>

            #include "test/platforms/metal_test.h"
            #include "test/type_helpers.h"
        """

    def class_preamble(self, fixture_name):
        return f"""
         private:
          struct Impl {{
            MetalTest* test;
          }};
          std::unique_ptr<Impl> impl_;

         protected:
          {fixture_name}() : impl_(std::make_unique<Impl>()) {{}}
          ~{fixture_name}() override {{}}

          void SetUp() override {{
            impl_->test = [[MetalTest alloc] initWithDevice:MTLCreateSystemDefaultDevice()];
          }}
          void TearDown() override {{
            [impl_->test release];
          }}
        """

    def framework_specific_args(self):
        return "impl_->test, @(kernel_name)"


if __name__ == "__main__":
    index, tu = parse(args.template)

    if tu.diagnostics:
        print(f"\nParsing template {args.template} failed with errors:")
        for d in tu.diagnostics:
            print(f"    {repr(d)}")
        print()
        raise ValueError("Non-zero compiler diagnostics, chickening out")

    if args.backend == "cuda":
        CudaTestFixture().generate(tu, args.out)
    elif args.backend == "hip":
        HipTestFixture().generate(tu, args.out)
    elif args.backend == "openmp":
        OpenmpTestFixture().generate(tu, args.out)
    elif args.backend == "opencl":
        OpenclTestFixture().generate(tu, args.out)
    elif args.backend == "metal":
        MetalTestFixture().generate(tu, args.out)
    elif args.backend == "sycl":
        SyclTestFixture().generate(tu, args.out)
    else:
        raise ValueError(f"Unknown backend provided: {args.backend}")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyvelox.pyvelox as pv
import unittest


class TestVeloxTypes(unittest.TestCase):
    def test_types(self):
        # Ensure we support all the basic types
        self.assertTrue(isinstance(pv.BooleanType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.IntegerType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.BigintType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.SmallintType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.TinyintType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.RealType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.DoubleType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.TimestampType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.VarcharType(), pv.VeloxType))
        self.assertTrue(isinstance(pv.VarbinaryType(), pv.VeloxType))

        # Complex types
        self.assertTrue(isinstance(pv.ArrayType(pv.BooleanType()), pv.VeloxType))
        self.assertTrue(
            isinstance(pv.MapType(pv.VarcharType(), pv.VarbinaryType()), pv.VeloxType)
        )
        self.assertTrue(
            isinstance(pv.RowType(["c0"], [pv.BooleanType()]), pv.VeloxType)
        )

    def test_complex_types(self):
        arrayType = pv.ArrayType(pv.BigintType())
        self.assertEqual(arrayType.element_type(), pv.BigintType())

        mapType = pv.MapType(pv.VarcharType(), pv.VarbinaryType())
        self.assertEqual(mapType.key_type(), pv.VarcharType())
        self.assertEqual(mapType.value_type(), pv.VarbinaryType())

        rowType = pv.RowType(
            ["c0", "c1", "c2"], [pv.BooleanType(), pv.BigintType(), pv.VarcharType()]
        )
        self.assertEqual(rowType.size(), 3)
        self.assertEqual(rowType.child_at(0), pv.BooleanType())
        self.assertEqual(rowType.find_child("c1"), pv.BigintType())
        self.assertEqual(rowType.get_child_idx("c1"), 1)
        self.assertEqual(rowType.name_of(1), "c1")
        self.assertEqual(rowType.names(), ["c0", "c1", "c2"])



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import distutils.command.build
import distutils.command.clean
import os
import re
import shutil
import subprocess
import sys
from pathlib import Path

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

ROOT_DIR = Path(__file__).parent.resolve()

with open("pyvelox/README.md") as f:
    readme = f.read()


# Override build directory
class BuildCommand(distutils.command.build.build):
    def initialize_options(self):
        distutils.command.build.build.initialize_options(self)
        self.build_base = "_build"


def _get_version():
    version = open("./version.txt").read().strip()
    version = re.sub("#.*\n?", "", version, flags=re.MULTILINE)
    sha = "Unknown"
    try:
        sha = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=str(ROOT_DIR))
            .decode("ascii")
            .strip()
        )
    except Exception:
        pass

    if os.getenv("BUILD_VERSION"):
        version = os.getenv("BUILD_VERSION")
    elif sha != "Unknown":
        version += "+" + sha[:7]

    return version, sha


def _export_version(version, sha):
    version_path = ROOT_DIR / "pyvelox" / "version.py"
    with open(version_path, "w") as f:
        f.write("__version__ = '{}'\n".format(version))
        f.write("git_version = {}\n".format(repr(sha)))


VERSION, SHA = _get_version()
_export_version(VERSION, SHA)

print("-- Building version " + VERSION)


class clean(distutils.command.clean.clean):
    def run(self):
        # Run default behavior first
        distutils.command.clean.clean.run(self)

        # Remove pyvelox extension
        for path in (ROOT_DIR / "pyvelox").glob("**/*.so"):
            print(f"removing '{path}'")
            path.unlink()
        # Remove build directory
        build_dirs = [
            ROOT_DIR / "_build",
        ]
        for path in build_dirs:
            if path.exists():
                print(f"removing '{path}' (and everything under it)")
                shutil.rmtree(str(path), ignore_errors=True)


# Based off of
# https://github.com/pytorch/audio/blob/2c8aad97fc8d7647ee8b2df2de9312cce0355ef6/build_tools/setup_helpers/extension.py#L46
class CMakeBuild(build_ext):
    def run(self):
        try:
            subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError("CMake is not available.")
        super().run()

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        # Allow using a pre-built Velox library (for CI and development) e.g. 'VELOX_BUILD_DIR=_build/velox/debug'
        # The build in question must have been built with 'VELOX_BUILD_PYTHON_PACKAGE=ON' and the same python version.
        if "VELOX_BUILD_DIR" in os.environ:
            velox_dir = os.path.abspath(os.environ["VELOX_BUILD_DIR"])

            if not os.path.isdir(extdir):
                os.symlink(velox_dir, os.path.dirname(extdir), target_is_directory=True)

            print(f"Using pre-built Velox library from {velox_dir}")
            return

        # required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        if "DEBUG" in os.environ:
            cfg = "Debug" if os.environ["DEBUG"] == "1" else "MinSizeRel"
        else:
            cfg = "Debug" if self.debug else "MinSizeRel"

        exec_path = sys.executable

        cmake_args = [
            "-DVELOX_BUILD_TESTING=OFF",
            "-DVELOX_MONO_LIBRARY=ON",
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
            f"-DCMAKE_BUILD_TYPE={cfg}",
            f"-DCMAKE_INSTALL_PREFIX={extdir}",
            "-DCMAKE_VERBOSE_MAKEFILE=ON",
            "-DVELOX_BUILD_PYTHON_PACKAGE=ON",
            f"-DPYTHON_EXECUTABLE={exec_path} ",
        ]
        build_args = []

        # Default to Ninja
        if "CMAKE_GENERATOR" not in os.environ:
            cmake_args += ["-GNinja"]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += ["-j{}".format(self.parallel)]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(
            ["cmake", str(ROOT_DIR)] + cmake_args,
            cwd=self.build_temp,
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


setup(
    name="pyvelox",
    version=VERSION,
    description="Python bindings and extensions for Velox",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/facebookincubator/velox",
    author="Meta",
    author_email="velox@fb.com",
    license="Apache License 2.0",
    install_requires=[
        "cffi",
        "typing",
        "tabulate",
        "typing-inspect",
        "pyarrow",
    ],
    extras_require={"tests": ["pyarrow"]},
    python_requires=">=3.9",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    packages=find_packages() + find_packages(where="./test"),
    zip_safe=False,
    ext_modules=[Extension(name="pyvelox.pyvelox", sources=[])],
    cmdclass={"build_ext": CMakeBuild, "clean": clean, "build": BuildCommand},
)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
import uuid
from os.path import join, splitext
from pathlib import Path
from typing import Any, Dict, List

from benchadapt import BenchmarkResult
from benchadapt.adapters import BenchmarkAdapter


class BinarySizeAdapter(BenchmarkAdapter):
    """
    Adapter to track build artifact sizes in conbench.
    Expects the `size_file` to be formatted like this:
    <size in bytes> <path/to/binary|arbitrary_name>

    Suite meta data will be library, object or executable
    based on file ending.
    """

    size_file: Path

    def __init__(
        self,
        command: List[str],
        size_file: str,
        build_type: str,
        result_fields_override: Dict[str, Any] = {},
        result_fields_append: Dict[str, Any] = {},
    ) -> None:
        self.size_file = Path(size_file)
        self.build_type = build_type
        super().__init__(command, result_fields_override, result_fields_append)

    def _transform_results(self) -> List[BenchmarkResult]:
        results = []

        batch_id = uuid.uuid4().hex
        with open(self.size_file, "r") as file:
            sizes = [line.strip() for line in file]

        if not sizes:
            raise ValueError("'size_file' is empty!")

        for line in sizes:
            size, path = line.split(maxsplit=1)
            path = path.strip()
            _, ext = splitext(path)
            if ext in [".so", ".a"]:
                suite = "library"
            elif ext == ".o":
                suite = "object"
            else:
                suite = "executable"

            parsed_size = BenchmarkResult(
                run_reason="merge",
                batch_id=batch_id,
                stats={
                    "data": [size],
                    "unit": "B",
                    "iterations": 1,
                },
                tags={
                    "name": path,
                    "suite": suite,
                    "source": f"{self.build_type}_build_metrics_size",
                },
                info={},
                context={"benchmark_language": "C++"},
            )
            results.append(parsed_size)

        return results


class NinjaLogAdapter(BenchmarkAdapter):
    """
    Adapter to extract compile and link times from a .ninja_log.
    Will calculate aggregates for total, compile and link time.
    Suite metadata will be set based on binary ending to object, library or executable.

    Only files in paths beginning with velox/ will be tracked to avoid dependencies.
    """

    ninja_log: Path

    def __init__(
        self,
        command: List[str],
        ninja_log: str,
        build_type: str,
        result_fields_override: Dict[str, Any] = {},
        result_fields_append: Dict[str, Any] = {},
    ) -> None:
        self.ninja_log = Path(ninja_log)
        self.build_type = build_type
        super().__init__(command, result_fields_override, result_fields_append)

    def _transform_results(self) -> List[BenchmarkResult]:
        results = []

        batch_id = uuid.uuid4().hex
        with open(self.ninja_log, "r") as file:
            log_lines = [line.strip() for line in file]

        if not log_lines[0].startswith("# ninja log v"):
            raise ValueError("Malformed Ninja log found!")
        else:
            del log_lines[0]

        ms2sec = lambda x: x / 1000
        get_epoch = lambda l: int(l.split()[2])
        totals = {
            "link_time": 0,
            "compile_time": 0,
            "total_time": 0,
        }

        for line in log_lines:
            start, end, epoch, object_path, _ = line.split()
            start = int(start)
            end = int(end)
            duration = ms2sec(end - start)

            # Mono build places library in lib/
            if not object_path.startswith(("velox", "lib/libvelox")):
                continue

            _, ext = splitext(object_path)
            if ext in [".so", ".a"] or not ext:
                totals["link_time"] += duration
                suite = "linking"
            elif ext == ".o":
                totals["compile_time"] += duration
                suite = "compiling"
            else:
                print(f"Unkown file type found: {object_path}")
                print("Skipping...")
                continue

            time_result = BenchmarkResult(
                run_reason="merge",
                batch_id=batch_id,
                stats={
                    "data": [duration],
                    "unit": "s",
                    "iterations": 1,
                },
                tags={
                    "name": object_path,
                    "suite": suite,
                    "source": f"{self.build_type}_build_metrics_time",
                },
                info={},
                context={"benchmark_language": "C++"},
            )
            results.append(time_result)

        totals["total_time"] = totals["link_time"] + totals["compile_time"]
        for total_name, total in totals.items():
            total_result = BenchmarkResult(
                run_reason="merge",
                batch_id=batch_id,
                stats={
                    "data": [total],
                    "unit": "s",
                    "iterations": 1,
                },
                tags={
                    "name": total_name,
                    "suite": "total",
                    "source": f"{self.build_type}_build_metrics_time",
                },
                info={},
                context={"benchmark_language": "C++"},
            )
            results.append(total_result)

        return results


# find velox -type f -name '*.o' -exec ls -l -BB {} \; | awk '{print $5, $9}' |  sed 's|CMakeFiles/.*dir/||g' > /tmp/object-size


def upload(args):
    print("Uploading Build Metrics")
    pr_number = int(args.pr_number) if args.pr_number else None
    run_reason = "pull request" if pr_number else "commit"
    run_name = f"{run_reason}: {args.sha}"
    sizes = BinarySizeAdapter(
        command=["true"],
        size_file=join(args.base_path, args.size_file),
        build_type=args.build_type,
        result_fields_override={
            "run_id": args.run_id,
            "run_name": run_name,
            "run_reason": run_reason,
            "github": {
                "repository": "https://github.com/facebookincubator/velox",
                "pr_number": pr_number,
                "commit": args.sha,
            },
        },
    )
    sizes()

    times = NinjaLogAdapter(
        command=["true"],
        ninja_log=join(args.base_path, args.ninja_log),
        build_type=args.build_type,
        result_fields_override={
            "run_id": args.run_id,
            "run_name": run_name,
            "run_reason": run_reason,
            "github": {
                "repository": "https://github.com/facebookincubator/velox",
                "pr_number": pr_number,
                "commit": args.sha,
            },
        },
    )
    times()


def parse_args():
    parser = argparse.ArgumentParser(description="Velox Build Metric Utility.")
    parser.set_defaults(func=lambda _: parser.print_help())

    subparsers = parser.add_subparsers(help="Please specify one of the subcommands.")

    upload_parser = subparsers.add_parser(
        "upload", help="Parse and upload build metrics"
    )
    upload_parser.set_defaults(func=upload)
    upload_parser.add_argument(
        "--ninja_log", default=".ninja_log", help="Name of the ninja log file."
    )
    upload_parser.add_argument(
        "--size_file",
        default="object_sizes",
        help="Name of the file containing size information.",
    )
    upload_parser.add_argument(
        "--build_type",
        required=True,
        help="Type of build results come from, e.g. debug or release",
    )
    upload_parser.add_argument(
        "--run_id",
        required=True,
        help="A Conbench run ID unique to this build.",
    )
    upload_parser.add_argument(
        "--sha",
        required=True,
        help="HEAD sha for the result upload to conbench.",
    )
    upload_parser.add_argument(
        "--pr_number",
        default=0,
        help="PR number for the result upload to conbench.",
    )
    upload_parser.add_argument(
        "base_path",
        help="Path in which the .ninja_log and sizes_file are found.",
    )

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    args.func(args)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from typing import List


class Vector:
    def size(self) -> int: ...
    def print_all(self) -> str: ...
    def print_detailed(self) -> str: ...
    def summarize_to_text(self) -> str: ...
    def null_count(self) -> int: ...
    def is_null_at(self, idx: int) -> bool: ...



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import pyvelox.pyvelox as pv


class TestFunctionSignatures(unittest.TestCase):
    def test_clear_signatures(self):
        pv.clear_signatures()
        signatures = pv.get_function_signatures()
        self.assertEqual(len(signatures), 0)

    def test_get_signatures(self):
        pv.register_presto_signatures()
        presto_signatures = pv.get_function_signatures()
        self.assertTrue(len(presto_signatures) > 0)

        pv.clear_signatures()
        pv.register_spark_signatures()
        spark_signatures = pv.get_function_signatures()
        self.assertTrue(len(spark_signatures) > 0)

    def test_function_signature(self):
        pv.clear_signatures()
        pv.register_presto_signatures()
        presto_signatures = pv.get_function_signatures()

        concat_signatures = presto_signatures["concat"]
        self.assertTrue(len(concat_signatures) > 0)
        # Array functions are registered first, then string functions.
        concat_signatures.sort(key=lambda sig: sig.__str__())
        self.assertEqual(str(concat_signatures[0].return_type()), "array(__user_T1)")
        self.assertEqual(
            str(concat_signatures[0]),
            "(__user_T1,array(__user_T1)) -> array(__user_T1)",
        )
        self.assertEqual(str(concat_signatures[-1].return_type()), "varchar")
        self.assertEqual(str(concat_signatures[-1]), "(varchar,varchar...) -> varchar")

    def test_function_prefix(self):
        pv.clear_signatures()
        pv.register_presto_signatures("foo")
        presto_signatures = pv.get_function_signatures()

        concat_signatures = presto_signatures["fooconcat"]
        self.assertTrue(len(concat_signatures) > 0)

        pv.clear_signatures()
        pv.register_spark_signatures("bar")
        spark_signatures = pv.get_function_signatures()

        concat_signatures = spark_signatures["barconcat"]
        self.assertTrue(len(concat_signatures) > 0)

    def test_aggregate_signatures(self):
        pv.clear_aggregate_signatures()

        pv.register_presto_aggregate_signatures()
        presto_agg_signatures = pv.get_aggregate_function_signatures()

        min_signatures = presto_agg_signatures["min"]
        self.assertTrue(len(min_signatures) > 0)

        max_signatures = presto_agg_signatures["max"]
        self.assertTrue(len(max_signatures) > 0)

        pv.clear_aggregate_signatures()

        pv.register_spark_aggregate_signatures()
        spark_agg_signatures = pv.get_aggregate_function_signatures()
        self.assertTrue(len(spark_agg_signatures) > 0)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

import unittest
import pyarrow

from velox.py.arrow import to_velox, to_arrow
from velox.py.type import DOUBLE


class TestPyVeloxVector(unittest.TestCase):
    def test_vector_size(self):
        vector = to_velox(pyarrow.array([1, 2, 3, 4, 5, 6]))
        self.assertEqual(len(vector), 6)
        self.assertEqual(vector.size(), 6)

    def test_vector_nulls(self):
        data = [1, 2, None, 4, 5, 6, None, 9, 10]
        vector = to_velox(pyarrow.array(data))
        self.assertEqual(vector.null_count(), 2)

        self.assertTrue(vector.is_null_at(2))
        self.assertTrue(vector.is_null_at(6))
        self.assertFalse(vector.is_null_at(0))
        self.assertFalse(vector.is_null_at(3))
        self.assertFalse(vector.is_null_at(8))

    def test_vector_compare(self):
        data = [1, 2, 3, 4, 5, 6, 7, 9, 10]
        vector1 = to_velox(pyarrow.array(data))
        vector2 = to_velox(to_arrow(vector1))

        # First compare each element.
        for i in range(len(data)):
            self.assertEqual(vector1.compare(vector2, i, i), 0)

        # Then the entire objects at once (__eq__).
        self.assertEqual(vector1, vector2)

        # Failures.
        vector3 = to_velox(pyarrow.array([1, 2, 3, 4, 5]))
        self.assertNotEqual(vector1, vector3)

        vector4 = to_velox(pyarrow.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))
        self.assertNotEqual(vector1, vector4)

    def test_vector_print(self):
        vector = to_velox(pyarrow.array([1, 2]))

        expected_header = "[FLAT BIGINT: 2 elements, no nulls]"
        self.assertEqual(str(vector), expected_header)
        self.assertTrue(vector.print_detailed().startswith(expected_header))

        self.assertEqual(vector.print_all(), "0: 1\n1: 2")
        self.assertEqual(vector.summarize_to_text(), "BIGINT 2 rows FLAT 16B\n")

    def test_vector_complex(self):
        # Array/lists.
        data = [[1, 2, 3], [4, None, 6], None, [7]]
        vector = to_velox(pyarrow.array(data))
        self.assertEqual(str(vector), "[ARRAY ARRAY<BIGINT>: 4 elements, 1 nulls]")

        # Maps.
        data = [[{"key": "a", "value": 1}, {"key": "b", "value": 2}]]
        map_type = pyarrow.map_(pyarrow.string(), pyarrow.int32())
        vector = to_velox(pyarrow.array(data, type=map_type))
        self.assertEqual(
            str(vector), "[MAP MAP<VARCHAR,INTEGER>: 1 elements, no nulls]"
        )

        # Row/structs.
        struct_type = pyarrow.struct(
            [("name", pyarrow.string()), ("age", pyarrow.int32())]
        )
        data = [{"name": "John", "age": 23}, {"name": "Mike", "age": 32}]
        vector = to_velox(pyarrow.array(data, type=struct_type))
        self.assertEqual(
            str(vector), "[ROW ROW<name:VARCHAR,age:INTEGER>: 2 elements, no nulls]"
        )

        # Validate children.
        self.assertEqual(
            str(vector.child_at(0)), "[FLAT VARCHAR: 2 elements, no nulls]"
        )
        self.assertEqual(
            str(vector.child_at(1)), "[FLAT INTEGER: 2 elements, no nulls]"
        )

    def test_vector_type(self):
        data = [1.9, 2.45]
        vector = to_velox(pyarrow.array(data))
        self.assertEqual(vector.type(), DOUBLE())
